#include <cstdio>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/soundcard.h>

#include "papu.hh"
#include "clock.h"

bisqpapu::bisqpapu() : Rate(48000)
{
    fd = 1;
    Rate = 44100;
    
    /*
    fd = open("/dev/dsp", O_WRONLY);
    if(fp)
    {
        int Bits = 16, Stereo = 0;
        ioctl(fd, SNDCTL_DSP_SAMPLESIZE, &Bits);
        ioctl(fd, SNDCTL_DSP_STEREO,     &Stereo);
        if(ioctl(fd, SNDCTL_DSP_SPEED, &Rate) < 0)
        {
            Rate = 44100;
            if(ioctl(fd, SNDCTL_DSP_SPEED, &Rate) < 0)
            {
                Rate = 22050;
                if(ioctl(fd, SNDCTL_DSP_SPEED, &Rate) < 0)
                {
                    Rate = 11025;
                    if(ioctl(fd, SNDCTL_DSP_SPEED, &Rate) < 0)
                    {
                        Rate = 8000;
                        ioctl(fd, SNDCTL_DSP_SPEED, &Rate);
                    }
                }
            }
        }
    }
    else
    {
        Rate = 44100;
        char cmdbuf[128];
        
        sprintf(cmdbuf, "esdcat -m -r %u", Rate);
        
        fp = popen(cmdbuf, "w");
    }
    */
    
    counter = 0;
    speed = double(Clock->CPUclockDivider * Rate) / Clock->ClockHz;
}

bisqpapu::~bisqpapu()
{
}

bool bisqpapu::write(const fWORD addr, fBYTE value)
{
    switch(addr)
    {
        case 0x4000 ... 0x4003: fprintf(stderr, "\33[1m1\33[m");chan1.write(addr&3, value); return true;
//        case 0x4004 ... 0x4007: fprintf(stderr, "\33[1m2\33[m");chan2.write(addr&3, value); return true;
//        case 0x4008 ... 0x400B: fprintf(stderr, "\33[1m3\33[m");chan3.write(addr&3, value); return true;
//        case 0x400C ... 0x400F: fprintf(stderr, "\33[1m4\33[m");chan4.write(addr&3, value); return true;
//        case 0x4010 ... 0x4013: fprintf(stderr, "\33[1m5\33[m");chan5.write(addr&3, value); return true;
        case 0x4015:
            if(value & 0x01)chan1.enable();else chan1.disable();
            if(value & 0x02)chan2.enable();else chan2.disable();
            if(value & 0x04)chan3.enable();else chan3.disable();
            if(value & 0x08)chan4.enable();else chan4.disable();
            if(value & 0x10)chan5.enable();else chan5.disable();
            // &32, &64 and &128 ignored.
            return true;
    }
    return false;
}
bool bisqpapu::read(const fWORD addr, BYTE &value)
{
    switch(addr)
    {
        case 0x4000 ... 0x4013:
            // Complete. These are write-only.
            value = 0x40;
            return true;
        case 0x4015:
        	fprintf(stderr, "\33[1;32mpAPU status: ");
            if(!chan1.lengthunit_silent())value |= 0x01;
            if(!chan2.lengthunit_silent())value |= 0x02;
            if(!chan3.lengthunit_silent())value |= 0x04;
            if(!chan4.lengthunit_silent())value |= 0x08;
            if(!chan5.silent())value |= 0x10;
            if(chan5.irqflag)value |= 0x80;
            fprintf(stderr, "%02Xn\33[m\n", value);
            return true;
    }
    return false;
}

void bisqpapu::mixbytes(unsigned count)
{
    static const BYTE dutyvalues[4] = {2,4,8,12};
    static const BYTE triangledata[32]=
    {13,14,15,15, 14,13,12,11, 10, 9, 8, 7,  6, 5, 4, 3,
      2, 1, 0, 0,  1, 2, 3, 4,  5, 6, 7, 8,  9,10,11,12};
    
    // chan1 = +-15, chan2 = +-15
    // chan3 = 
    
    int val1 = chan1.silent() ? 0 : ((chan1.dutycounter&15) < dutyvalues[chan1.dutytype] ? -1 : 1) * chan1.getvol();
    int val2 = chan2.silent() ? 0 : ((chan2.dutycounter&15) < dutyvalues[chan2.dutytype] ? -1 : 1) * chan2.getvol();
    int val3 = chan3.silent() ? 0 : triangledata[chan3.trianglecounter&31];
    int val4 = chan4.silent() ? 0 : chan4.output * chan4.getvol();
    int val5 = chan5.silent() ? 0 : chan5.deltacounter;
    
    // val1, val2:   -15 .. +15
    // val3: 0..15
    // val4: 0..15
    // val5: 0..127
    
    register short d = ((val1+val2-val3+val4) // range: -15-15-15+0 .. 15+15-0+15 = -45..+45
                        * 3                   // range: -135..+135
                        + val5)               // range: -135..+262
                        * (32767/500);
    
    register BYTE lo = d&255;
    register BYTE hi = d>>8;
    register fWORD b = count*2;
    BYTE buf[b];        
    for(register fWORD a=0; a<b; )
    {
        buf[a++] = lo;
        buf[a++] = hi;
    }
    std::write(fd, buf, b);
}

bool bisqpapu::tick()
{
    // This function runs Clock->ClockHz / Clock->CPUclockDivider
    // times per second. For NTSC, it is 1.7897725 MHz.

    chan1.tick();
    chan2.tick();
    chan3.tick();
    chan4.tick();
    bool irq = chan5.tick();

    register unsigned bytecount = (unsigned)(counter += speed);
    if(bytecount)
    {
        mixbytes(bytecount);
        counter -= bytecount;
    }

    return irq;
}
